#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools
import inspect
import re

import netaddr
from os_ken.lib.packet import ether_types
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import netutils
from oslo_utils import strutils
from oslo_utils import uuidutils
from webob import exc

from neutron_lib._i18n import _
from neutron_lib.api import converters
from neutron_lib import constants
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory


LOG = logging.getLogger(__name__)

# Used by range check to indicate no limit for a bound.
UNLIMITED = None

# Note: In order to ensure that the MAC address is unicast the first byte
# must be even.
MAC_PATTERN = "^%s[aceACE02468](:%s{2}){5}$" % (constants.HEX_ELEM,
                                                constants.HEX_ELEM)


def _verify_dict_keys(expected_keys, target_dict, strict=True):
    """Verify expected keys in a dictionary.

    :param expected_keys: A list of keys expected to be present.
    :param target_dict: The dictionary which should be verified.
    :param strict: Specifies whether additional keys are allowed to be present.
    :returns: None if the expected keys are found. Otherwise a human readable
        message indicating why the validation failed.
    """
    if not isinstance(target_dict, dict):
        msg = (_("Invalid input. '%(target_dict)s' must be a dictionary "
                 "with keys: %(expected_keys)s") %
               {'target_dict': target_dict, 'expected_keys': expected_keys})
        LOG.debug(msg)
        return msg

    expected_keys = set(expected_keys)
    provided_keys = set(target_dict.keys())

    predicate = expected_keys.__eq__ if strict else expected_keys.issubset

    if not predicate(provided_keys):
        msg = (_("Validation of dictionary's keys failed. "
                 "Expected keys: %(expected_keys)s "
                 "Provided keys: %(provided_keys)s") %
               {'expected_keys': expected_keys,
                'provided_keys': provided_keys})
        LOG.debug(msg)
        return msg


def _collect_duplicates(data_list):
    """Collects duplicate items from a list and returns them.

    :param data_list: A list of items to check for duplicates. The list may
        include dict items.
    :returns: A list of items that are duplicates in data_list. If no
        duplicates are found, the returned list is empty.
    """
    seen = []
    dups = []
    for datum in data_list:
        if datum in seen and datum not in dups:
            dups.append(datum)
            continue
        seen.append(datum)
    return dups


def is_attr_set(attribute):
    """Determine if an attribute value is set.

    :param attribute: The attribute value to check.
    :returns: False if the attribute value is None or ATTR_NOT_SPECIFIED,
        otherwise True.
    """
    return not (attribute is None or
                attribute is constants.ATTR_NOT_SPECIFIED)


def _validate_list_of_items(item_validator, data, *args, **kwargs):
    if not isinstance(data, list):
        msg = _("'%s' is not a list") % data
        return msg

    dups = _collect_duplicates(data)

    if dups:
        msg = _("Duplicate items in the list: '%s'") % ', '.join(
            [str(d) for d in dups])
        return msg

    for item in data:
        msg = item_validator(item, *args, **kwargs)
        if msg:
            return msg


def _validate_list_of_items_non_empty(item_validator, data, *args, **kwargs):
    res = _validate_list_of_items(item_validator, data, *args, **kwargs)

    if res is not None:
        return res

    if len(data) == 0:
        msg = _("List should not be empty")
        return msg


def validate_values(data, valid_values=None, valid_values_display=None):
    """Validate that the provided 'data' is within 'valid_values'.

    :param data: The data to check within valid_values.
    :param valid_values: A collection of values that 'data' must be in to be
        valid. The collection can be any type that supports the 'in' operation.
    :param valid_values_display: A string to display that describes the valid
        values. This string is only displayed when an invalid value is
        encountered.
        If no string is provided, the string "valid_values" will be used.
    :returns: The message to return if data not in valid_values.
    :raises: TypeError if the values for 'data' or 'valid_values' are not
        compatible for comparison or doesn't have __contains__.
        If TypeError is raised this is considered a programming error and the
        inputs (data) and (valid_values) must be checked so this is never
        raised on validation.
    """

    # If valid_values is not specified we don't check against it.
    if valid_values is None:
        return

    # Check if we can use 'in' to find membership of data in valid_values
    contains = getattr(valid_values, "__contains__", None)
    if callable(contains):
        try:
            if data not in valid_values:
                valid_values_display = valid_values_display or 'valid_values'
                msg = (_("%(data)s is not in %(valid_values)s") %
                       {'data': data, 'valid_values': valid_values_display})
                LOG.debug(msg)
                return msg
        except TypeError as e:
            # This is a programming error
            msg = (_("'data' of type '%(typedata)s' and 'valid_values' "
                     "of type '%(typevalues)s' are not "
                     "compatible for comparison") %
                   {'typedata': type(data),
                    'typevalues': type(valid_values)})
            raise TypeError(msg) from e
    else:
        # This is a programming error
        msg = (_("'valid_values' does not support membership operations"))
        raise TypeError(msg)


def validate_not_empty_string_or_none(data, max_len=None):
    """Validate data is a non-empty string or None.

    :param data: The data to validate.
    :param max_len: An optional cap on the str length to validate.
    :returns: None if the data string is not None and is not an empty string,
        otherwise a human readable message as to why the string data is
        invalid.
    """
    if data is not None:
        return validate_not_empty_string(data, max_len=max_len)


def validate_not_empty_string(data, max_len=None):
    """Validate data is a non-empty/non-blank string.

    :param data: The data to validate.
    :param max_len: An optional cap on the length of the string data.
    :returns: None if the data is non-empty/non-blank, otherwise a human
        readable string message indicating why validation failed.
    """
    msg = validate_string(data, max_len=max_len)
    if msg:
        return msg
    if not data.strip():
        msg = _("'%s' Blank strings are not permitted") % data
        LOG.debug(msg)
        return msg


def validate_string_or_none(data, max_len=None):
    """Validate data is a string or None.

    :param data: The data to validate.
    :param max_len: An optional cap on the length of the string data.
    :returns: None if the data is None or a valid string, otherwise a human
        readable message indicating why validation failed.
    """
    if data is not None:
        return validate_string(data, max_len=max_len)


def validate_string(data, max_len=None):
    """Validate data is a string object optionally capping it length.

    :param data: The data to validate.
    :param max_len: An optional cap on the length of the string.
    :returns: None if the data is a valid string type and (optionally) within
        the given max_len. Otherwise a human readable message indicating why
        the data is invalid.
    """
    if not isinstance(data, str):
        msg = _("'%s' is not a valid string") % data
        LOG.debug(msg)
        return msg

    if max_len is not None and len(data) > max_len:
        msg = (_("'%(data)s' exceeds maximum length of %(max_len)s") %
               {'data': data, 'max_len': max_len})
        LOG.debug(msg)
        return msg


_validate_list_of_unique_strings = functools.partial(_validate_list_of_items,
                                                     validate_string)


# NOTE(boden): stubbed out for docstring comments.
def validate_list_of_unique_strings(data, max_len=None):
    """Validate data is a list of unique strings.

    :param data: The data to validate.
    :param max_len: An optional cap on the length of the string.
    :returns: None if the data is a list of non-empty/non-blank strings,
        otherwise a human readable message indicating why validation failed.
    """
    return _validate_list_of_unique_strings(data, max_len=max_len)


def validate_oneline_not_empty_string(data, max_len=None):
    """Validate data is a non-empty string without newline character.

    :param data: The data to validate.
    :param max_len: An optional cap on the length of the string.
    :returns: None if the data is a string without newline character,
        otherwise a human readable message indicating why validation failed.
    """
    msg = validate_not_empty_string(data, max_len=max_len)
    if msg:
        return msg
    if len(data.splitlines()) > 1:
        msg = _("Multi-line string is not allowed: '%s'") % data
        LOG.debug(msg)
        return msg


def validate_oneline_not_empty_string_or_none(data, max_len=None):
    """Validate data is a non-empty string without newline character or None.

    :param data: The data to validate.
    :param max_len: An optional cap on the length of the string data.
    :returns: None if the data is None or a valid string without newline
        character, otherwise a human readable message indicating why
        validation failed.
    """
    if data is not None:
        return validate_oneline_not_empty_string(data, max_len=max_len)


def validate_boolean(data, valid_values=None):
    """Validate data is a python bool compatible object.

    :param data: The data to validate.
    :param valid_values: Not used!
    :return: None if the value can be converted to a bool, otherwise a
        human readable message indicating why data is invalid.
    """
    try:
        strutils.bool_from_string(data, strict=True)
    except ValueError:
        msg = _("'%s' is not a valid boolean value") % data
        LOG.debug(msg)
        return msg


def validate_integer(data, valid_values=None):
    """This function validates if the data is an integer.

    It checks both number or string provided to validate it's an
    integer and returns a message with the error if it's not

    :param data: The string or number to validate as integer.
    :param valid_values: values to limit the 'data' to.
    :returns: None if data is an integer, otherwise a human readable message
        indicating why validation failed..
    """

    if valid_values is not None:
        msg = validate_values(data=data, valid_values=valid_values)
        if msg:
            return msg

    msg = _("'%s' is not an integer") % data
    try:
        fl_n = float(data)
        int_n = int(data)
    except (ValueError, TypeError, OverflowError):
        LOG.debug(msg)
        return msg

    # Fail test if non equal or boolean
    if fl_n != int_n:
        LOG.debug(msg)
        return msg
    elif isinstance(data, bool):
        msg = _("'%s' is not an integer:boolean") % data
        LOG.debug(msg)
        return msg


def validate_range(data, valid_values=None):
    """Check that integer value is within a range provided.

    Test is inclusive. Allows either limit to be ignored, to allow
    checking ranges where only the lower or upper limit matter.
    It is expected that the limits provided are valid integers or
    the value None.

    :param data: The data to validate.
    :param valid_values: A list of 2 elements where element 0 is the min
        value the int data can have and element 1 is the max.
    :returns: None if the data is a valid int in the given range, otherwise
        a human readable message as to why validation failed.
    """

    min_value = valid_values[0]
    max_value = valid_values[1]
    try:
        data = int(data)
    except (ValueError, TypeError):
        msg = _("'%s' is not an integer") % data
        LOG.debug(msg)
        return msg
    if min_value is not UNLIMITED and data < min_value:
        msg = _("'%(data)s' is too small - must be at least "
                "'%(limit)d'") % {'data': data, 'limit': min_value}
        LOG.debug(msg)
        return msg
    if max_value is not UNLIMITED and data > max_value:
        msg = _("'%(data)s' is too large - must be no larger than "
                "'%(limit)d'") % {'data': data, 'limit': max_value}
        LOG.debug(msg)
        return msg


def validate_range_or_none(data, valid_values=None):
    """Check that the provided value is none or a ranged integer"""

    if data is not None:
        return validate_range(data, valid_values)


def validate_no_whitespace(data):
    """Validates that input has no whitespace.

    :param data: The data to validate. Must be a python string type suitable
        for searching via regex.
    :returns: The data itself.
    :raises InvalidInput: If the data contains whitespace.
    """
    if re.search(r'\s', data):
        msg = _("'%s' contains whitespace") % data
        LOG.debug(msg)
        raise n_exc.InvalidInput(error_message=msg)
    return data


def validate_mac_address(data, valid_values=None):
    """Validate data is a MAC address.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if the data is a valid MAC address, otherwise a human
        readable message as to why validation failed.
    """
    try:
        valid_mac = netaddr.valid_mac(validate_no_whitespace(data))
    except Exception:
        valid_mac = False

    if valid_mac:
        valid_mac = (netaddr.EUI(data) not in
                     map(netaddr.EUI, constants.INVALID_MAC_ADDRESSES))
    # TODO(arosen): The code in this file should be refactored
    # so it catches the correct exceptions. validate_no_whitespace
    # raises AttributeError if data is None.
    if not valid_mac:
        msg = _("'%s' is not a valid MAC address") % data
        LOG.debug(msg)
        return msg


def validate_mac_address_or_none(data, valid_values=None):
    """Validate data is a MAC address if the data isn't None.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if the data is None or a valid MAC address, otherwise
        a human readable message indicating why validation failed.
    """
    if data is not None:
        return validate_mac_address(data, valid_values)


def validate_ip_address(data, valid_values=None):
    """Validate data is an IP address.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is an IP address, otherwise a human readable
        message indicating why data isn't an IP address.
    """
    msg = None
    try:
        # netaddr.core.ZEROFILL is only applicable to IPv4.
        # it will remove leading zeros from IPv4 address octets.
        ip = netaddr.IPAddress(validate_no_whitespace(data),
                               flags=netaddr.core.ZEROFILL)
        # The followings are quick checks for IPv6 (has ':') and
        # IPv4.  (has 3 periods like 'xx.xx.xx.xx')
        # NOTE(yamamoto): netaddr uses libraries provided by the underlying
        # platform to convert addresses.  For example, inet_aton(3).
        # Some platforms, including NetBSD and OS X, have inet_aton
        # implementation which accepts more varying forms of addresses than
        # we want to accept here.  The following check is to reject such
        # addresses.  For Example:
        #   >>> netaddr.IPAddress('1' * 59)
        #   IPAddress('199.28.113.199')
        #   >>> netaddr.IPAddress(str(int('1' * 59) & 0xffffffff))
        #   IPAddress('199.28.113.199')
        #   >>>
        if ':' not in data and data.count('.') != 3:
            msg = _("'%s' is not a valid IP address") % data
        # A leading '0' in IPv4 address may be interpreted as an octal number,
        # e.g. 011 octal is 9 decimal. Since there is no standard saying
        # whether IP address with leading '0's should be interpreted as octal
        # or decimal, hence we reject leading '0's to avoid ambiguity.
        elif ip.version == 4 and str(ip) != data:
            msg = _("'%(data)s' is not an accepted IP address, "
                    "'%(ip)s' is recommended") % {"data": data, "ip": ip}
    except Exception:
        msg = _("'%s' is not a valid IP address") % data
    if msg:
        LOG.debug(msg)
    return msg


def _validate_any_key_spec(data, key_specs=None):
    """Validate a dict matches at least 1 key spec.

    :param data: The dict to validate.
    :param key_specs: An iterable collection of key spec dicts used to validate
        data.
    :returns: None if at least 1 of the key_specs matches data, otherwise
        a message is returned indicating data could not be matched with any
        of the key_specs.
    """
    for spec in key_specs:
        if validate_dict(data, spec) is None:
            return None
    msg = _("No valid key specs matched for: %s") % data
    LOG.debug(msg)
    return msg


def validate_any_key_specs_or_none(data, key_specs=None):
    """Validate each dict in a list matches at least 1 key_spec.

    :param data: The list of dicts to validate.
    :param key_specs: An iterable collection of key spec dicts that is used
        to check each dict in data.
    :returns: None.
    :raises InvalidInput: If any of the dicts in data do not match at least
        1 of the key_specs given.
    """
    if data is None:
        return

    def dict_validator(data_dict):
        msg = _validate_any_key_spec(data_dict, key_specs=key_specs)
        if msg:
            raise n_exc.InvalidInput(error_message=msg)

    msg = _validate_list_of_items(dict_validator, data)
    if msg:
        raise n_exc.InvalidInput(error_message=msg)


def validate_ip_pools(data, valid_values=None):
    """Validate that start and end IP addresses are present.

    In addition to this the IP addresses will also be validated.

    :param data: The data to validate. Must be a list-like structure of
        IP pool dicts that each have a 'start' and 'end' key value.
    :param valid_values: Not used!
    :returns: None if data is a valid list of IP pools, otherwise a message
        indicating why the data is invalid.
    """
    if not isinstance(data, list):
        msg = _("Invalid data format for IP pool: '%s'") % data
        LOG.debug(msg)
        return msg

    expected_keys = ['start', 'end']
    for ip_pool in data:
        msg = _verify_dict_keys(expected_keys, ip_pool)
        if msg:
            return msg
        for k in expected_keys:
            msg = validate_ip_address(ip_pool[k])
            if msg:
                return msg


def validate_fixed_ips(data, valid_values=None):
    """Validate data is a list of fixed IP dicts.

    In addition this function validates the ip_address and subnet_id
    if present in each fixed IP dict.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is a valid list of fixed IP dicts. Otherwise a
        human readable message is returned indicating why validation failed.
    """
    if not isinstance(data, list):
        msg = _("Invalid data format for fixed IP: '%s'") % data
        LOG.debug(msg)
        return msg

    ips = []
    for fixed_ip in data:
        if not isinstance(fixed_ip, dict):
            msg = _("Invalid data format for fixed IP: '%s'") % fixed_ip
            LOG.debug(msg)
            return msg
        if 'ip_address' in fixed_ip:
            # Ensure that duplicate entries are not set - just checking IP
            # suffices. Duplicate subnet_id's are legitimate.
            fixed_ip_address = fixed_ip['ip_address']
            if fixed_ip_address in ips:
                msg = _("Duplicate IP address '%s'") % fixed_ip_address
                LOG.debug(msg)
            else:
                msg = validate_ip_address(fixed_ip_address)
            if msg:
                return msg
            ips.append(fixed_ip_address)
        if 'subnet_id' in fixed_ip:
            msg = validate_uuid(fixed_ip['subnet_id'])
            if msg:
                return msg


def validate_nameservers(data, valid_values=None):
    """Validate a list of unique IP addresses.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is a list of valid IP addresses, otherwise
        a human readable message is returned indicating why validation failed.
    """
    if not hasattr(data, '__iter__'):
        msg = _("Invalid data format for nameserver: '%s'") % data
        LOG.debug(msg)
        return msg

    hosts = []
    for host in data:
        # This must be an IP address only
        msg = validate_ip_address(host)
        if msg:
            msg = _("'%(host)s' is not a valid nameserver. %(msg)s") % {
                'host': host, 'msg': msg}
            LOG.debug(msg)
            return msg
        if host in hosts:
            msg = _("Duplicate nameserver '%s'") % host
            LOG.debug(msg)
            return msg
        hosts.append(host)


def validate_hostroutes(data, valid_values=None):
    """Validate a list of unique host route dicts.

    :param data: The data to validate. To be valid it must be a list like
        structure of host route dicts, each containing 'destination' and
        'nexthop' key values.
    :param valid_values: A list of valid keys or None
    :returns: None if data is a valid list of unique host route dicts,
        otherwise a human readable message indicating why validation failed.
    """
    if not isinstance(data, list):
        msg = _("Invalid data format for hostroute: '%s'") % data
        LOG.debug(msg)
        return msg

    hostroutes = []
    for hostroute in data:
        expected_keys = ['destination', 'nexthop']
        if hostroute and 'bfd_monitor_id' in hostroute:
            expected_keys.append('bfd_monitor_id')
        msg = _verify_dict_keys(expected_keys, hostroute)
        if msg:
            return msg
        msg = validate_route_cidr(hostroute['destination'])
        if msg:
            return msg
        msg = validate_ip_address(hostroute['nexthop'])
        if msg:
            return msg
        if hostroute in hostroutes:
            msg = _("Duplicate hostroute '%s'") % hostroute
            LOG.debug(msg)
            return msg
        if 'bfd_monitor_id' in hostroute:
            msg = validate_uuid(hostroute['bfd_monitor_id'])
            if msg:
                return msg
        hostroutes.append(hostroute)


def validate_ip_address_or_none(data, valid_values=None):
    """Validate data is an IP address or None.

    :param data: The data to validate.
    :param valid_values: An optional list of values data may take on.
    :return: None if data is None or a valid IP address, otherwise a
        human readable message indicating why the data is invalid.
    """
    if data is not None:
        return validate_ip_address(data, valid_values)


def validate_ip_or_subnet_or_none(data, valid_values=None):
    """Validate data is an IP address, a valid IP subnet string, or None.

    :param data: The data to validate.
    :param valid_values: Not used!
    :return: None if data is None or a valid IP address or a valid IP subnet,
        otherwise a human readable message indicating why the data is neither
        an IP address nor IP subnet.
    """
    msg_ip = validate_ip_address_or_none(data)
    msg_subnet = validate_subnet_or_none(data)
    if msg_ip is not None and msg_subnet is not None:
        return _("'%(data)s' is neither a valid IP address, nor "
                 "is it a valid IP subnet") % {'data': data}


def validate_subnet(data, valid_values=None):
    """Validate data is an IP network subnet string.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is valid IP network address. Otherwise a human
        readable message as to why data is invalid.
    """
    msg = None
    try:
        net = netaddr.IPNetwork(validate_no_whitespace(data))
        if '/' not in data or (net.version == 4 and str(net) != data):
            msg = _("'%(data)s' isn't a recognized IP subnet cidr,"
                    " '%(cidr)s' is recommended") % {"data": data,
                                                     "cidr": net.cidr}
        else:
            return
    except Exception:
        msg = _("'%s' is not a valid IP subnet") % data
    if msg:
        LOG.debug(msg)
    return msg


def validate_route_cidr(data, valid_values=None):
    """Validate data is a proper CIDR string.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is valid CIDR. Otherwise a human
              readable message as to why data is invalid.
    """
    msg = None
    try:
        net = netaddr.IPNetwork(validate_no_whitespace(data))
        if '/' not in data or (net.network != net.ip):
            msg = _("'%(data)s' is not a recognized CIDR,"
                    " '%(cidr)s' is recommended") % {"data": data,
                                                     "cidr": net.cidr}
        elif net.is_loopback():
            msg = _("'%(data)s' is not a routable CIDR") % {"data": data}
        else:
            return
    except Exception:
        msg = _("'%s' is not a valid CIDR") % data
    if msg:
        LOG.debug(msg)
    return msg


def validate_subnet_or_none(data, valid_values=None):
    """Validate data is a valid subnet address string or None.

    :param data: The data to validate.
    :param valid_values: The optional list of values data may take on.
    :returns: None if data is None or a valid subnet, otherwise a human
        readable message as to why data is invalid.
    """
    if data is not None:
        return validate_subnet(data, valid_values)


_validate_subnet_list = functools.partial(_validate_list_of_items,
                                          validate_subnet)


# NOTE(boden): subbed out for docstring comments.
def validate_subnet_list(data, valid_values=None):
    """Validate data is a list of subnet dicts.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is a valid list of subnet dicts, otherwise a human
        readable message as to why the data is invalid.
    """
    return _validate_subnet_list(data, valid_values)


def validate_subnet_list_or_none(data, key_specs=None):
    """Validate data is a list of subnet dicts or None.

    :param data: The data to validate.
    :param key_specs: Not used!
    :returns: None if data is None or a valid list of subnet dicts, otherwise
        a human readable message as to why the data is invalid.
    """
    if data is not None:
        return validate_subnet_list(data, key_specs)


def validate_regex(data, valid_values=None):
    """Validate data is matched against a regex.

    :param data: The data to validate.
    :param valid_values: The regular expression to use with re.match on
        the data.
    :returns: None if data contains matches for valid_values, otherwise a
        human readable message as to why data is invalid.
    """
    try:
        if re.match(valid_values, data):
            return
    except TypeError:
        pass

    msg = _("'%s' is not a valid input") % data
    LOG.debug(msg)
    return msg


def validate_regex_or_none(data, valid_values=None):
    """Validate data is None or matched against a regex.

    :param data: The data to validate.
    :param valid_values: The regular expression to use with re.match on
        the data.
    :returns: None if data is None or contains matches for valid_values,
        otherwise a human readable message as to why data is invalid.
    """
    if data is not None:
        return validate_regex(data, valid_values)


def validate_list_of_regex_or_none(data, valid_values=None):
    """Validate data is None or a list of items matching regex.

    :param data: A list of data to validate.
    :param valid_values: The regular expression to use with re.match on
        each element of the data.
    :returns: None if data is None or contains matches for valid_values,
        otherwise a human readable message as to why data is invalid.
    """
    if data is not None:
        return _validate_list_of_items(validate_regex, data, valid_values)


def validate_subnetpool_id(data, valid_values=None):
    """Validate data is valid subnet pool ID.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is a valid subnet pool ID, otherwise a
        human readable message as to why it's invalid.
    """
    if data != constants.IPV6_PD_POOL_ID:
        return validate_uuid_or_none(data, valid_values)


def validate_subnetpool_id_or_none(data, valid_values=None):
    """Validate data is valid subnet pool ID or None.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is a valid subnet pool ID or None, otherwise a
        human readable message as to why it's invalid.
    """
    if data is not None:
        return validate_subnetpool_id(data, valid_values)


def validate_uuid(data, valid_values=None):
    """Validate data is UUID like.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is UUID like in form, otherwise a human readable
        message indicating why data is invalid.
    """
    if not uuidutils.is_uuid_like(data):
        msg = _("'%s' is not a valid UUID") % data
        LOG.debug(msg)
        return msg


def validate_uuid_or_none(data, valid_values=None):
    """Validate data is UUID like or None.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is UUID like in form or None, otherwise a human
        readable message indicating why data is invalid.
    """
    if data is not None:
        return validate_uuid(data)


_validate_uuid_list = functools.partial(_validate_list_of_items,
                                        validate_uuid)


# NOTE(boden): subbed out for docstring comments.
def validate_uuid_list(data, valid_values=None):
    """Validate data is a list of UUID like values.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is an iterable that contains valid UUID values,
        otherwise a message is returned indicating why validation failed.
    """
    return _validate_uuid_list(data, valid_values)


def validate_uuid_list_non_empty(data, valid_values=None):
    """Validate data is a non-empty list of UUID like values.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is a non-empty iterable that contains valid UUID
        values, otherwise a message is returned indicating why validation
        failed.
    """
    return _validate_list_of_items_non_empty(validate_uuid, data)


def _extract_validator(key_validator):
    # Find validator function in key validation spec
    #
    # TODO(salv-orlando): Structure of dict attributes should be improved
    # to avoid iterating over items
    for (k, v) in key_validator.items():
        if k.startswith('type:'):
            (validator_name, validator_params) = (k, v)
            try:
                return (validator_name,
                        validators[validator_name],
                        validator_params)
            except KeyError as e:
                raise UndefinedValidator(validator_name) from e
    return None, None, None


def _validate_dict_item(key, key_validator, data):
    # Find conversion function, if any, and apply it
    conv_func = key_validator.get('convert_to')
    if conv_func:
        data[key] = conv_func(data.get(key))
    try:
        dummy_, val_func, val_params = _extract_validator(key_validator)
        if val_func:
            return val_func(data.get(key), val_params)
        # NOTE(tmorin): here we silently omit to validate a key for which
        # no type validator has been defined
    except UndefinedValidator as e:
        # NOTE(tmorin): Should we really return an API error on such
        # an issue. Wouldn't InternalServer error be more natural ?
        LOG.debug(e.message)
        return e.message


def validate_dict(data, key_specs=None):
    """Validate data is a dict optionally containing a specific set of keys.

    :param data: The data to validate.
    :param key_specs: The optional list of keys that must be contained in
        data.
    :returns: None if data is a dict and (optionally) contains only key_specs.
        Otherwise a human readable message is returned indicating why data is
        not valid.
    """
    if not isinstance(data, dict):
        msg = _("'%s' is not a dictionary") % data
        LOG.debug(msg)
        return msg
    # Do not perform any further validation, if no constraints are supplied
    if not key_specs:
        return

    # Check whether all required keys are present
    required_keys = [key for key, spec in key_specs.items()
                     if spec.get('required')]

    if required_keys:
        msg = _verify_dict_keys(required_keys, data, False)
        if msg:
            return msg

    # Check whether unexpected keys are supplied in data
    unexpected_keys = [key for key in data if key not in key_specs]
    if unexpected_keys:
        msg = _("Unexpected keys supplied: %s") % ', '.join(unexpected_keys)
        LOG.debug(msg)
        return msg

    # Perform validation and conversion of all values
    # according to the specifications.
    for key, key_validator in [(k, v) for k, v in key_specs.items()
                               if k in data]:
        msg = _validate_dict_item(key, key_validator, data)
        if msg:
            return msg


def validate_dict_or_none(data, key_specs=None):
    """Validate data is None or a dict containing a specific set of keys.

    :param data: The data to validate.
    :param key_specs: The optional list of keys that must be contained in
        data.
    :returns: None if data is None or a dict  that (optionally) contains
        all key_specs. Otherwise a human readable message is returned
        indicating why data is not valid.
    """
    if data is not None:
        return validate_dict(data, key_specs)


def validate_dict_or_empty(data, key_specs=None):
    """Validate data is {} or a dict containing a specific set of keys.

    :param data: The data to validate.
    :param key_specs: The optional list of keys that must be contained in
        data.
    :returns: None if data is {} or a dict (optionally) containing
        only key_specs. Otherwise a human readable message is returned
        indicating why data is not valid.
    """
    if data != {}:
        return validate_dict(data, key_specs)


def validate_dict_or_nodata(data, key_specs=None):
    """Validate no data or a dict containing a specific set of keys.

    :param data: The data to validate. May be None.
    :param key_specs: The optional list of keys that must be contained in
        data.
    :returns: None if no data/empty dict or a dict and (optionally) contains
        all key_specs. Otherwise a human readable message is returned
        indicating why data is not valid.
    """
    if data:
        return validate_dict(data, key_specs)


def validate_non_negative(data, valid_values=None):
    """Validate data is a positive int.

    :param data: The data to validate
    :param valid_values: Not used!
    :returns: None if data is an int and is positive, otherwise a human
        readable message as to why data is invalid.
    """
    try:
        data = int(data)
    except (ValueError, TypeError):
        msg = _("'%s' is not an integer") % data
        LOG.debug(msg)
        return msg

    if data < 0:
        msg = _("'%s' should be non-negative") % data
        LOG.debug(msg)
        return msg


def validate_port_range_or_none(data, valid_values=None):
    """Validate data is a range of TCP/UDP port numbers

    :param data: The data to validate
    :param valid_values: Not used!
    :returns: None if data is an int between 0 and 65535, or two ints between 0
        and 65535 with a colon between them, otherwise a human readable message
        as to why data is invalid.
    """
    if data is None:
        return
    data = str(data)
    ports = data.split(':')
    if len(ports) > 2:
        msg = _("Port range must be two integers separated by a colon")
        LOG.debug(msg)
        return msg
    for p in ports:
        if len(p) == 0:
            msg = _("Port range must be two integers separated by a colon")
            LOG.debug(msg)
            return msg
        if not netutils.is_valid_port(p):
            msg = _("Invalid port: %s") % p
            LOG.debug(msg)
            return msg
    if len(ports) > 1 and int(ports[0]) > int(ports[1]):
        msg = _("First port in a port range must be lower than the second "
                "port")
        LOG.debug(msg)
        return msg


def validate_subports(data, valid_values=None):
    """Validate data is a list of subnet port dicts.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if data is a list of subport dicts each with a unique valid
        port_id, segmentation_id and segmentation_type. Otherwise a human
        readable message is returned indicating why the data is invalid.
    """
    if not isinstance(data, list):
        msg = _("Invalid data format for subports: '%s' is not a list") % data
        LOG.debug(msg)
        return msg

    subport_ids = set()
    segmentations = collections.defaultdict(set)
    for subport in data:
        if not isinstance(subport, dict):
            msg = _("Invalid data format for subport: "
                    "'%s' is not a dict") % subport
            LOG.debug(msg)
            return msg

        # Expect a non duplicated and valid port_id for the subport
        if 'port_id' not in subport:
            msg = _("A valid port UUID must be specified")
            LOG.debug(msg)
            return msg
        elif validate_uuid(subport["port_id"]):
            msg = _("Invalid UUID for subport: '%s'") % subport["port_id"]
            return msg
        elif subport["port_id"] in subport_ids:
            msg = _("Non unique UUID for subport: '%s'") % subport["port_id"]
            return msg
        subport_ids.add(subport["port_id"])

        # Validate that both segmentation ID and segmentation type are
        # specified, and that the client does not duplicate segmentation
        # IDs (unless it is explicitly asked to inherit segmentation
        # details from the underlying subport's network).
        segmentation_type = subport.get("segmentation_type")
        if segmentation_type == 'inherit':
            return
        segmentation_id = subport.get("segmentation_id")
        if ((not segmentation_type or segmentation_id is None) and
                len(subport) > 1):
            msg = _("Invalid subport details '%s': missing segmentation "
                    "information. Must specify both segmentation_id and "
                    "segmentation_type") % subport
            LOG.debug(msg)
            return msg
        if segmentation_id in segmentations.get(segmentation_type, []):
            msg = _("Segmentation ID '%(seg_id)s' for '%(subport)s' is not "
                    "unique") % {"seg_id": segmentation_id,
                                 "subport": subport["port_id"]}
            LOG.debug(msg)
            return msg
        if segmentation_id is not None:
            segmentations[segmentation_type].add(segmentation_id)


def validate_service_plugin_type(data, valid_values=None):
    """Validates data is a valid service plugin.

    :param data: The service plugin type to validate.
    :param valid_values: Not used.
    :returns: None if data is a valid service plugin known to the plugin
        directory.
    :raises: InvalidServiceType if data is not a service known by the
        plugin directory.
    """
    if not directory.get_plugin(data):
        raise n_exc.InvalidServiceType(service_type=data)


def validate_subnet_service_types(service_types, valid_values=None):
    if service_types:
        if not isinstance(service_types, list):
            raise exc.HTTPBadRequest(
                _("Subnet service types must be a list."))

        # Include standard prefixes
        prefixes = list(constants.DEVICE_OWNER_PREFIXES)
        prefixes += constants.DEVICE_OWNER_COMPUTE_PREFIX

        for service_type in service_types:
            if not isinstance(service_type, str):
                raise n_exc.InvalidInputSubnetServiceType(
                    service_type=service_type)
            if not service_type.startswith(tuple(prefixes)):
                raise n_exc.InvalidSubnetServiceType(service_type=service_type)


def validate_ethertype(ethertype, valid_values=None):
    """Validates ethertype is a valid ethertype.

    :param ethertype: Ethertype to validate.
    :returns: None if data is a valid ethertype.  Otherwise a human-readable
        message indicating that the data is not a valid ethertype.
    """

    if cfg.CONF.sg_filter_ethertypes:
        os_ken_ethertypes = ether_types.__dict__
        try:
            # Use the ethertype constants from os_ken with the special
            # exception of 'IPv4', which appears as 'ETH_TYPE_IP' in os_ken.
            # Be case-insensitive.
            ethertype_str = str(ethertype).upper()
            if ethertype_str == "IPV4":
                ethertype_str = "IP"
            ethertype_name = 'ETH_TYPE_' + ethertype_str
            ethertype = os_ken_ethertypes.get(ethertype_name, ethertype)
        except TypeError:
            # Value of ethertype cannot be coerced into a string, like None
            pass
        try:
            if isinstance(ethertype, str):
                ethertype = int(ethertype, 16)
            if (isinstance(ethertype, int) and
                    constants.ETHERTYPE_MIN <= ethertype <=
                    constants.ETHERTYPE_MAX):
                return None
        except ValueError:
            pass
        msg = _("Ethertype %s is not a two octet hexadecimal "
                "number or ethertype name.") % ethertype
        LOG.debug(msg)
        return msg
    else:
        if ethertype in constants.VALID_ETHERTYPES:
            return None
        valids = ','.join(map(str, constants.VALID_ETHERTYPES))
        msg = _("Ethertype %(ethertype)s is not a valid ethertype, must be "
                "one of %(valid_ethertypes)s.") % {'ethertype': ethertype,
                                                   'valid_ethertypes': valids}
        LOG.debug(msg)
        return msg


def validate_external_gw_info(data, valid_values=None):
    """Validate data is an external_gateway_info.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if valid, error string otherwise.
    """
    return validate_dict_or_nodata(
        data,
        key_specs={
            'network_id': {'type:uuid': None, 'required': True},
            'external_fixed_ips': {'type:fixed_ips': None, 'required': False},
            'enable_snat': {'type:boolean': None, 'required': False,
                            'convert_to': converters.convert_to_boolean},
        }
    )


def validate_external_gw_info_list(data, valid_values=None):
    """Validate data is a list of external_gateway_info.

    :param data: The data to validate.
    :param valid_values: Not used!
    :returns: None if valid, error string otherwise.
    """
    if data is not None:
        return _validate_list_of_items(validate_external_gw_info, data)


# Dictionary that maintains a list of validation functions
validators = {'type:dict': validate_dict,
              'type:dict_or_none': validate_dict_or_none,
              'type:dict_or_empty': validate_dict_or_empty,
              'type:dict_or_nodata': validate_dict_or_nodata,
              'type:ethertype': validate_ethertype,
              'type:fixed_ips': validate_fixed_ips,
              'type:hostroutes': validate_hostroutes,
              'type:ip_address': validate_ip_address,
              'type:ip_address_or_none': validate_ip_address_or_none,
              'type:ip_or_subnet_or_none': validate_ip_or_subnet_or_none,
              'type:ip_pools': validate_ip_pools,
              'type:list_of_regex_or_none': validate_list_of_regex_or_none,
              'type:mac_address': validate_mac_address,
              'type:mac_address_or_none': validate_mac_address_or_none,
              'type:nameservers': validate_nameservers,
              'type:non_negative': validate_non_negative,
              'type:port_range': validate_port_range_or_none,
              'type:range': validate_range,
              'type:range_or_none': validate_range_or_none,
              'type:regex': validate_regex,
              'type:regex_or_none': validate_regex_or_none,
              'type:string': validate_string,
              'type:string_or_none': validate_string_or_none,
              'type:not_empty_string': validate_not_empty_string,
              'type:not_empty_string_or_none':
              validate_not_empty_string_or_none,
              'type:oneline_not_empty_string':
              validate_oneline_not_empty_string,
              'type:oneline_not_empty_string_or_none':
              validate_oneline_not_empty_string_or_none,
              'type:subnet': validate_subnet,
              'type:subnet_list': validate_subnet_list,
              'type:subnet_or_none': validate_subnet_or_none,
              'type:subnetpool_id': validate_subnetpool_id,
              'type:subnetpool_id_or_none': validate_subnetpool_id_or_none,
              'type:subports': validate_subports,
              'type:uuid': validate_uuid,
              'type:uuid_or_none': validate_uuid_or_none,
              'type:uuid_list': validate_uuid_list,
              'type:uuid_list_non_empty': validate_uuid_list_non_empty,
              'type:values': validate_values,
              'type:boolean': validate_boolean,
              'type:integer': validate_integer,
              'type:list_of_unique_strings': validate_list_of_unique_strings,
              'type:list_of_any_key_specs_or_none':
                  validate_any_key_specs_or_none,
              'type:service_plugin_type': validate_service_plugin_type,
              'type:list_of_subnets_or_none': validate_subnet_list_or_none,
              'type:list_of_subnet_service_types':
                  validate_subnet_service_types,
              'type:external_gw_info': validate_external_gw_info,
              'type:external_gw_info_list': validate_external_gw_info_list,
              }


class UndefinedValidator(Exception):

    def __init__(self, validator_name):
        self.validator_name = validator_name
        self.message = _("Validator '%s' does not exist") % self.validator_name


def _to_validation_type(validation_type):
    return (validation_type
            if validation_type.startswith('type:')
            else 'type:' + validation_type)


def get_validator(validation_type, default=None):
    """Get a registered validator by type.

    :param validation_type: The type to retrieve the validator for.
    :param default: A default value to return if the validator is
        not registered.
    :return: The validator if registered, otherwise the default value.
    """
    return validators.get(_to_validation_type(validation_type), default)


def add_validator(validation_type, validator):
    """Dynamically add a validator.

    This can be used by clients to add their own, private validators, rather
    than directly modifying the data structure. The clients can NOT modify
    existing validators.
    """
    key = _to_validation_type(validation_type)
    if key in validators:
        # NOTE(boden): imp.load_source() forces module reinitialization that
        # can lead to validator redefinition from the same call site
        if inspect.getsource(validator) != inspect.getsource(validators[key]):
            msg = _("Validator type %s is already defined") % validation_type
            raise KeyError(msg)
        return
    validators[key] = validator



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib import constants
from neutron_lib.db import constants as db_const


NAME = 'Neutron Port'
ALIAS = 'port'
DESCRIPTION = "Network port abstraction"

UPDATED_TIMESTAMP = "2012-01-01T10:00:00-00:00"

RESOURCE_NAME = 'port'
COLLECTION_NAME = 'ports'
COLLECTION_NAME_BULK = 'ports_bulk'

PORT_MAC_ADDRESS = 'mac_address'

RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_filter': True,
               'is_sort_key': True,
               'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True, 'default': '',
                 'validate': {
                     'type:string': db_const.NAME_FIELD_SIZE},
                 'is_filter': True,
                 'is_sort_key': True,
                 'is_visible': True},
        'network_id': {'allow_post': True, 'allow_put': False,
                       'required_by_policy': True,
                       'validate': {'type:uuid': None},
                       'is_filter': True,
                       'is_sort_key': True,
                       'is_visible': True},
        'admin_state_up': {'allow_post': True, 'allow_put': True,
                           'default': True,
                           'convert_to': converters.convert_to_boolean,
                           'is_filter': True,
                           'is_sort_key': True,
                           'is_visible': True},
        PORT_MAC_ADDRESS: {'allow_post': True, 'allow_put': True,
                           'default': constants.ATTR_NOT_SPECIFIED,
                           'convert_to':
                               converters.convert_to_sanitized_mac_address,
                           'validate': {'type:mac_address': None},
                           'enforce_policy': True,
                           'is_filter': True,
                           'is_sort_key': True,
                           'is_visible': True},
        'fixed_ips': {'allow_post': True, 'allow_put': True,
                      'default': constants.ATTR_NOT_SPECIFIED,
                      'convert_list_to':
                          converters.convert_kvp_list_to_dict,
                      'validate': {'type:fixed_ips': None},
                      'enforce_policy': True,
                      'is_filter': True,
                      'is_visible': True},
        'device_id': {'allow_post': True, 'allow_put': True,
                      'validate': {
                          'type:string': db_const.DEVICE_ID_FIELD_SIZE},
                      'default': '',
                      'is_filter': True,
                      'is_sort_key': True,
                      'is_visible': True},
        'device_owner': {'allow_post': True, 'allow_put': True,
                         'validate': {
                             'type:string': db_const.DEVICE_OWNER_FIELD_SIZE},
                         'default': '', 'enforce_policy': True,
                         'is_filter': True,
                         'is_sort_key': True,
                         'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {
                          'type:string': db_const.PROJECT_ID_FIELD_SIZE},
                      'required_by_policy': True,
                      'is_filter': True,
                      'is_sort_key': True,
                      'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_filter': True,
                   'is_sort_key': True,
                   'is_visible': True},
    }
}

# This is a core resource so the following are not applicable.
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



# Copyright 2021 Ericsson Software Technology
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.api.definitions import l3
from neutron_lib.api.definitions import l3_ext_gw_mode


ALIAS = 'multiple-external-gateways'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Neutron L3 Multiple External Gateways'
API_PREFIX = ''
DESCRIPTION = 'Allow more than one External Gateways'
UPDATED_TIMESTAMP = '2021-04-01T00:00:00-00:00'
RESOURCE_NAME = l3.ROUTER
COLLECTION_NAME = l3.ROUTERS
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        'external_gateways': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
            'default': None,
            'validate': {'type:external_gw_info_list': None},
        },
    },
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = l3.ACTION_MAP
ACTION_MAP[l3.ROUTER].update({
    'add_external_gateways': 'PUT',
    'update_external_gateways': 'PUT',
    'remove_external_gateways': 'PUT',
})
REQUIRED_EXTENSIONS = [l3.ALIAS, l3_ext_gw_mode.ALIAS]
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



# Copyright (c) 2021 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import qos as qos_apidef
from neutron_lib.api.definitions import qos_pps_rule
from neutron_lib.services.qos import constants as qos_const
from neutron_lib.tests.unit.api.definitions import base


class QoSPPSRuleDefinitionTestCase(base.DefinitionBaseTestCase):
    extension_module = qos_pps_rule

    extension_resources = (qos_apidef.POLICIES,)
    extension_subresources = (qos_apidef.BANDWIDTH_LIMIT_RULES,
                              qos_apidef.DSCP_MARKING_RULES,
                              qos_apidef.MIN_BANDWIDTH_RULES,
                              qos_pps_rule.PACKET_RATE_LIMIT_RULES)
    extension_attributes = (qos_const.MAX_KPPS, qos_const.MAX_BURST_KPPS,
                            qos_const.DIRECTION, qos_const.MAX_BURST, 'type',
                            qos_const.DSCP_MARK, qos_const.MIN_KBPS, 'rules',
                            qos_const.MAX_KBPS, qos_const.QOS_POLICY_ID)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
NOTE: This module is a temporary shim until networking projects move to
      versioned objects at which point this module shouldn't be needed.
"""
from oslo_db.sqlalchemy import utils as sa_utils
from sqlalchemy.orm import lazyload
from sqlalchemy import sql, or_, and_

from neutron_lib._i18n import _
from neutron_lib.api import attributes
from neutron_lib import constants
from neutron_lib.db import utils as db_utils
from neutron_lib import exceptions as n_exc
from neutron_lib.objects import utils as obj_utils
from neutron_lib.utils import helpers


# Classes implementing extensions will register hooks into this dictionary
# for "augmenting" the "core way" of building a query for retrieving objects
# from a model class. Hooks are registered by invoking register_hook().
_model_query_hooks = {
    # model1 : {
    #              hook1: {
    #                         'query': query_hook,
    #                         'filter': filter_hook,
    #                         'result_filters': result_filters
    #              },
    #              hook2: {
    #                         'query': query_hook,
    #                         'filter': filter_hook,
    #                         'result_filters': result_filters
    #              },
    #              ...
    #          },
    # model2 : {
    #              hook1: {
    #                         'query': query_hook,
    #                         'filter': filter_hook,
    #                         'result_filters': result_filters
    #              },
    #              hook2: {
    #                         'query': query_hook,
    #                         'filter': filter_hook,
    #                         'result_filters': result_filters
    #              },
    #              ...
    #          },
    # ...
}


def register_hook(model, name, query_hook, filter_hook,
                  result_filters=None):
    """Register a hook to be invoked when a query is executed.

    Adds the hook components to the _model_query_hooks dict. Models are the
    keys of this dict, whereas the value is another dict mapping hook names
    to callables performing the hook.

    :param model: The DB Model that the hook applies to.
    :param name: A name for the hook.
    :param query_hook: The method to be called to augment the query.
    :param filter_hook: A method to be called to augment the query filter.
    :param result_filters: A Method to be called to filter the query result.
    :returns: None.
    """
    if callable(query_hook):
        query_hook = helpers.make_weak_ref(query_hook)
    if callable(filter_hook):
        filter_hook = helpers.make_weak_ref(filter_hook)
    if callable(result_filters):
        result_filters = helpers.make_weak_ref(result_filters)
    _model_query_hooks.setdefault(model, {})[name] = {
        'query': query_hook,
        'filter': filter_hook,
        'result_filters': result_filters
    }


def get_hooks(model):
    """Retrieve the model query hooks for a model.

    :param model: The DB Model to look up for query hooks.
    :returns: list of hooks
    """
    return _model_query_hooks.get(model, {}).values()


def query_with_hooks(context, model, field=None, lazy_fields=None):
    """Query with hooks using the said context and model.

    :param context: The context to use for the DB session.
    :param model: The model to query.
    :param field: The column.
    :param lazy_fields: list of fields for lazy loading
    :returns: The query with hooks applied to it.
    """
    if field:
        if hasattr(model, field):
            field = getattr(model, field)
        else:
            msg = _("'%s' is not supported as field") % field
            raise n_exc.InvalidInput(error_message=msg)
        query = context.session.query(field)
    else:
        query = context.session.query(model)
    # define basic filter condition for model query
    query_filter = None
    if db_utils.model_query_scope_is_project(context, model):
        if hasattr(model, 'rbac_entries'):
            query = query.outerjoin(model.rbac_entries)
            rbac_model = model.rbac_entries.property.mapper.class_
            query_filter = (
                (model.tenant_id == context.tenant_id) |
                (rbac_model.action.in_(
                    [constants.ACCESS_SHARED, constants.ACCESS_READONLY]) &
                 ((rbac_model.target_project == context.tenant_id) |
                  (rbac_model.target_project == '*'))))
        elif hasattr(model, 'shared'):
            query_filter = ((model.tenant_id == context.tenant_id) |
                            (model.shared == sql.true()))
        else:
            query_filter = (model.tenant_id == context.tenant_id)
    # Execute query hooks registered from mixins and plugins
    for hook in get_hooks(model):
        query_hook = helpers.resolve_ref(hook.get('query'))
        if query_hook:
            query = query_hook(context, model, query)

        filter_hook = helpers.resolve_ref(hook.get('filter'))
        if filter_hook:
            query_filter = filter_hook(context, model, query_filter)

    # NOTE(salvatore-orlando): 'if query_filter' will try to evaluate the
    # condition, raising an exception
    if query_filter is not None:
        query = query.filter(query_filter)

    if lazy_fields:
        for field in lazy_fields:
            query = query.options(lazyload(field))
    return query


def get_by_id(context, model, object_id, lazy_fields=None):
    """Query the said model with the given context for a specific object.

    :param context: The context to use in the query.
    :param model: The model to query.
    :param object_id: The ID of the object to query for.
    :param lazy_fields: list of fields for lazy loading
    :returns: The object with the give object_id for the said model.
    """
    query = query_with_hooks(context=context, model=model,
                             lazy_fields=lazy_fields)
    return query.filter(model.id == object_id).one()


def apply_filters(query, model, filters, context=None):
    """Apply filters to a query.

    :param query: The query to apply filters to.
    :param model: The model for the query.
    :param filters: The filters to apply.
    :param context: The context to use for the DB session.
    :returns: The query with filters applied to it.
    """
    if filters:
        for key, value in filters.items():
            column = getattr(model, key, None)
            # NOTE(kevinbenton): if column is a hybrid property that
            # references another expression, attempting to convert to
            # a boolean will fail so we must compare to None.
            # See "An Important Expression Language Gotcha" in:
            # docs.sqlalchemy.org/en/rel_0_9/changelog/migration_06.html
            if column is not None:
                if not value:
                    query = query.filter(sql.false())
                    return query
                if not hasattr(column, 'in_'):
                    # NOTE(ralonsoh): since SQLAlchemy==1.3.0, a column is an
                    # AssociationProxyInstance and inherits in_() method from
                    # ColumnOperators.
                    # association proxies don't support in_ so we have to
                    # do multiple equals matches
                    query = query.filter(
                        or_(*[column == v for v in value]))
                elif isinstance(value, obj_utils.FilterObj):
                    query = query.filter(value.filter(column))
                elif None in value:
                    # in_() operator does not support NULL element so we have
                    # to do multiple equals matches
                    query = query.filter(
                        or_(*[column == v for v in value]))
                else:
                    # NOTE(frickler): in_() isn't implemented for relations
                    # yet, let this pass so it can be handled by the
                    # result_filter hook
                    try:
                        query = query.filter(column.in_(value))
                    except NotImplementedError:
                        pass
            elif key == 'shared' and hasattr(model, 'rbac_entries'):
                # translate a filter on shared into a query against the
                # object's rbac entries
                rbac = model.rbac_entries.property.mapper.class_
                matches = [rbac.target_project == '*']
                if context:
                    matches.append(rbac.target_project == context.tenant_id)
                # any 'access_as_shared' records that match the
                # wildcard or requesting tenant
                is_shared = and_(rbac.action == constants.ACCESS_SHARED,
                                 or_(*matches))
                if not value[0]:
                    # NOTE(kevinbenton): we need to find objects that don't
                    # have an entry that matches the criteria above so
                    # we use a subquery to exclude them.
                    # We can't just filter the inverse of the query above
                    # because that will still give us a network shared to
                    # our tenant (or wildcard) if it's shared to another
                    # tenant.
                    # This is the column joining the table to rbac via
                    # the object_id. We can't just use model.id because
                    # subnets join on network.id so we have to inspect the
                    # relationship.
                    join_cols = model.rbac_entries.property.local_columns
                    oid_col = list(join_cols)[0]
                    is_shared = ~oid_col.in_(
                        query.session.query(rbac.object_id).filter(is_shared)
                    )
                elif (not context or
                      not db_utils.model_query_scope_is_project(
                          context, model)):
                    # we only want to join if we aren't using the subquery
                    # and if we aren't already joined because this is a
                    # scoped query
                    query = query.outerjoin(model.rbac_entries)
                query = query.filter(is_shared)
        for hook in get_hooks(model):
            result_filter = helpers.resolve_ref(
                hook.get('result_filters', None))
            if result_filter:
                query = result_filter(query, filters)
    return query


def get_collection_query(context, model, filters=None, sorts=None, limit=None,
                         marker_obj=None, page_reverse=False, field=None,
                         lazy_fields=None):
    """Get a collection query.

    :param context: The context to use for the DB session.
    :param model: The model to use.
    :param filters: The filters to apply in the query.
    :param sorts: The sort keys to use.
    :param limit: The limit associated with the query.
    :param marker_obj: The marker object if applicable.
    :param page_reverse: If reverse paging should be used.
    :param field: Column, in string format, from the "model"; the query will
                  return only this parameter instead of the full model columns.
    :param lazy_fields: list of fields for lazy loading
    :returns: A paginated query for the said model.
    """
    collection = query_with_hooks(context, model, field=field,
                                  lazy_fields=lazy_fields)
    collection = apply_filters(collection, model, filters, context)
    if sorts:
        sort_keys = db_utils.get_and_validate_sort_keys(sorts, model)
        sort_dirs = db_utils.get_sort_dirs(sorts, page_reverse)
        # we always want deterministic results for sorted queries
        # so add unique keys to limit queries when present.
        # (http://docs.sqlalchemy.org/en/latest/orm/
        #  loading_relationships.html#subqueryload-ordering)
        # (http://docs.sqlalchemy.org/en/latest/faq/
        #  ormconfiguration.html#faq-subqueryload-limit-sort)
        for k in _unique_keys(model):
            if k not in sort_keys:
                sort_keys.append(k)
                sort_dirs.append('asc')
        collection = sa_utils.paginate_query(collection, model, limit,
                                             marker=marker_obj,
                                             sort_keys=sort_keys,
                                             sort_dirs=sort_dirs)
    return collection


def _unique_keys(model):
    # just grab first set of unique keys and use them.
    # if model has no unqiue sets, 'paginate_query' will
    # warn if sorting is unstable
    uk_sets = sa_utils.get_unique_keys(model)
    return uk_sets[0] if uk_sets else []


def get_collection(context, model, dict_func,
                   filters=None, fields=None,
                   sorts=None, limit=None, marker_obj=None,
                   page_reverse=False, lazy_fields=None):
    """Get a collection for a said model.

    :param context: The context to use for the DB session.
    :param model: The model for the collection.
    :param dict_func: The function used to build the collection dict.
    :param filters: The filters to apply.
    :param fields: The fields to collect.
    :param sorts: The sort keys to use.
    :param limit: The limit for the query if applicable.
    :param marker_obj: The marker object if applicable.
    :param page_reverse: If reverse paging should be used.
    :param lazy_fields: list of fields for lazy loading
    :returns: A list of dicts where each dict is an object in the collection.
    """
    query = get_collection_query(context, model,
                                 filters=filters, sorts=sorts,
                                 limit=limit, marker_obj=marker_obj,
                                 page_reverse=page_reverse,
                                 lazy_fields=lazy_fields)
    items = [
        attributes.populate_project_info(
            dict_func(c, fields) if dict_func else c)
        for c in query
    ]
    if limit and page_reverse:
        items.reverse()
    return items


def get_values(context, model, field, filters=None):
    query = query_with_hooks(context, model, field=field)
    query = apply_filters(query, model, filters, context)
    return [c[0] for c in query]


def get_collection_count(context, model, filters=None, query_field=None):
    """Get the count for a specific collection.

    :param context: The context to use for the DB session.
    :param model: The model for the query.
    :param filters: The filters to apply.
    :param query_field: Column, in string format, from the "model"; the query
                        will return only this parameter instead of the full
                        model columns.
    :returns: The number of objects for said model with filters applied.
    """
    return get_collection_query(context, model, filters=filters,
                                field=query_field).count()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.api.validators import availability_zone as az_validator
from neutron_lib.db import constants as db_const
from neutron_lib import exceptions
from neutron_lib.tests import _base as base


class TestAvailabilityZoneValidator(base.BaseTestCase):

    @mock.patch.object(az_validator.validators,
                       'validate_list_of_unique_strings',
                       return_value='bad')
    def test__validate_availability_zone_hints_unique_strings(
            self, mock_unique_strs):
        self.assertEqual(
            'bad', az_validator._validate_availability_zone_hints(['a', 'a']))

    def test__validate_availability_zone_hints_excessive_len(self):
        self.assertRaisesRegex(
            exceptions.InvalidInput, 'Too many availability_zone_hints',
            az_validator._validate_availability_zone_hints,
            ['a' * (db_const.AZ_HINTS_DB_LEN + 1)])

    def test__validate_availability_zone_hints_valid_input(self):
        self.assertIsNone(
            az_validator._validate_availability_zone_hints(['a', 'b', 'c']))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api.definitions import l3
from neutron_lib.api import validators
from neutron_lib.api.validators import bfd as bfd_validator
from neutron_lib import constants
from neutron_lib.db import constants as db_const


ALIAS = 'bfd-monitor'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'BFD monitors for Neutron'
DESCRIPTION = "Provides support for BFD monitors"
UPDATED_TIMESTAMP = "2021-07-26T11:00:00-00:00"
BFD_MONITOR = 'bfd_monitor'
BFD_MONITORS = 'bfd_monitors'
BFD_SESSION_STATUS = 'bfd_session_status'

BFD_MODE_ASYNC = 'asynchronous'
BFD_MODE_DEMAND = 'demand'
BFD_MODE_ONE_ARM = 'one_arm_echo'

AUTH_TYPE_PWD = 'password'  # nosec
AUTH_TYPE_MD5 = 'MD5'
AUTH_TYPE_METIC_MD5 = 'MeticulousMD5'
AUTH_TYPE_SHA1 = 'SHA1'
AUTH_TYPE_METIC_SHA1 = 'MeticulousSHA1'

VALID_AUTH_TYPES = (AUTH_TYPE_PWD, AUTH_TYPE_MD5, AUTH_TYPE_METIC_MD5,
                    AUTH_TYPE_SHA1, AUTH_TYPE_METIC_SHA1)
VALID_MODES = (BFD_MODE_ASYNC, BFD_MODE_DEMAND, BFD_MODE_ONE_ARM)

validators.add_validator('bfd_mode_validator',
                         bfd_validator.validate_bfd_mode)
validators.add_validator('bfd_auth_type_validator',
                         bfd_validator.validate_bfd_auth_type)

RESOURCE_ATTRIBUTE_MAP = {
    BFD_MONITORS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True,
               'primary_key': True,
               'enforce_policy': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': db_const.NAME_FIELD_SIZE},
                 'default': '', 'is_filter': True, 'is_sort_key': True,
                 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'is_visible': True, 'default': '',
                        'validate': {
                            'type:string': db_const.DESCRIPTION_FIELD_SIZE}},
        'project_id': {'allow_post': True, 'allow_put': False,
                       'validate': {
                           'type:string': db_const.PROJECT_ID_FIELD_SIZE},
                       'required_by_policy': True,
                       'is_visible': True, 'enforce_policy': True},
        'mode': {'allow_post': True, 'allow_put': False,
                 'validate': {'type:bfd_mode_validator': VALID_MODES},
                 'default': BFD_MODE_ASYNC, 'is_filter': True,
                 'is_sort_key': True, 'is_visible': True},
        'dst_ip': {'allow_post': True, 'allow_put': False,
                   'validate': {'type:ip_address': None},
                   'is_sort_key': True, 'is_filter': True,
                   'is_visible': True, 'default': None,
                   'enforce_policy': True},
        'src_ip': {'allow_post': True, 'allow_put': False,
                   'validate': {'type:ip_address_or_none': None},
                   'is_sort_key': True, 'is_filter': True,
                   'is_visible': True, 'default': None,
                   'enforce_policy': True},
        'min_rx': {'allow_post': True, 'allow_put': True,
                   'validate': {'type:non_negative': None},
                   'convert_to': converters.convert_to_int,
                   'default': 1000,
                   'is_visible': True, 'enforce_policy': True},
        'min_tx': {'allow_post': True, 'allow_put': True,
                   'validate': {'type:non_negative': None},
                   'convert_to': converters.convert_to_int,
                   'default': 100,
                   'is_visible': True, 'enforce_policy': True},
        'multiplier': {'allow_post': True, 'allow_put': True,
                       'validate': {'type:non_negative': None},
                       'convert_to': converters.convert_to_int,
                       'default': 3,
                       'is_visible': True, 'enforce_policy': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_filter': True, 'is_sort_key': True,
                   'is_visible': True},
        'auth_type': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:bfd_auth_type_validator':
                                   VALID_AUTH_TYPES},
                      'default': constants.ATTR_NOT_SPECIFIED,
                      'is_visible': True},
        'auth_key': {'allow_post': True, 'allow_put': False,
                     'validate': {'type:dict_or_none': None},
                     'default': constants.ATTR_NOT_SPECIFIED,
                     'is_visible': True},
    },
}

SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {
    BFD_MONITOR: {
        'get_bfd_session_status': 'GET',
        'get_bfd_monitor_associations': 'GET',
    }
}
ACTION_STATUS = {}
REQUIRED_EXTENSIONS = [l3.ALIAS]
OPTIONAL_EXTENSIONS = []



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.api import converters
from neutron_lib import constants
from neutron_lib.db import constants as db_const
from neutron_lib.services.qos import constants as qos_consts


METERING_LABELS = 'metering_labels'
METERING_LABEL_RULES = 'metering_label_rules'

ALIAS = 'metering'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Neutron Metering'
API_PREFIX = ''
DESCRIPTION = 'Neutron Metering extension.'
UPDATED_TIMESTAMP = '2013-06-12T10:00:00-00:00'
RESOURCE_ATTRIBUTE_MAP = {
    METERING_LABELS: {
        'id': {
            'allow_post': False, 'allow_put': False,
            'is_visible': True,
            'is_filter': True,
            'is_sort_key': True,
            'primary_key': True
        },
        'name': {
            'allow_post': True, 'allow_put': False,
            'validate': {'type:string': db_const.NAME_FIELD_SIZE},
            'is_filter': True, 'is_sort_key': True,
            'is_visible': True, 'default': ''
        },
        'description': {
            'allow_post': True, 'allow_put': False,
            'validate': {
                'type:string': db_const.LONG_DESCRIPTION_FIELD_SIZE},
            'is_filter': True, 'is_sort_key': True,
            'is_visible': True, 'default': ''
        },
        'tenant_id': {
            'allow_post': True, 'allow_put': False,
            'required_by_policy': True,
            'validate': {
                'type:string': db_const.PROJECT_ID_FIELD_SIZE},
            'is_filter': True, 'is_sort_key': True,
            'is_visible': True
        },
        constants.SHARED: {
            'allow_post': True, 'allow_put': False,
            'is_filter': True, 'is_sort_key': True,
            'is_visible': True, 'default': False,
            'convert_to': converters.convert_to_boolean
        }
    },
    METERING_LABEL_RULES: {
        'id': {
            'allow_post': False, 'allow_put': False,
            'is_visible': True, 'is_filter': True,
            'is_sort_key': True, 'primary_key': True
        },
        'metering_label_id': {
            'allow_post': True, 'allow_put': False,
            'validate': {'type:uuid': None},
            'is_filter': True, 'is_sort_key': True,
            'is_visible': True, 'required_by_policy': True
        },
        qos_consts.DIRECTION: {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'is_filter': True,
            'is_sort_key': True,
            'validate': {'type:values': constants.VALID_DIRECTIONS}
        },
        'excluded': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'default': False,
            'is_filter': True, 'is_sort_key': True,
            'convert_to': converters.convert_to_boolean
        },
        'remote_ip_prefix': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'required_by_policy': True,
            'is_filter': True, 'is_sort_key': True,
            'validate': {'type:subnet': None}
        },
        'tenant_id': {
            'allow_post': True, 'allow_put': False,
            'required_by_policy': True,
            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
            'is_visible': True
        }
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.db import constants as db_const

# The alias of the extension.
ALIAS = 'logging'

# Whether or not this extension is simply signaling behavior to the user
# or it actively modifies the attribute map.
IS_SHIM_EXTENSION = False

# Whether the extension is marking the adoption of standardattr model for
# legacy resources, or introducing new standardattr attributes. False or
# None if the standardattr model is adopted since the introduction of
# resource extension.
# If this is True, the alias for the extension should be prefixed with
# 'standard-attr-'.
IS_STANDARD_ATTR_EXTENSION = False

# The name of the extension.
NAME = 'Logging API Extension'

# A prefix for API resources. An empty prefix means that the API is going
# to be exposed at the v2/ level as any other core resource.
API_PREFIX = '/log'

# The description of the extension.
DESCRIPTION = ("Provides a logging API for resources such as "
               "security group and firewall.")

# A timestamp of when the extension was introduced.
UPDATED_TIMESTAMP = "2017-01-01T10:00:00-00:00"

LOGS = 'logs'
LOG_TYPES = 'loggable_resources'

ACCEPT_EVENT = 'ACCEPT'
DROP_EVENT = 'DROP'
ALL_EVENT = 'ALL'
LOG_EVENTS = [ACCEPT_EVENT, DROP_EVENT, ALL_EVENT]

# Attribute Map
RESOURCE_ATTRIBUTE_MAP = {
    LOGS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True,
               'is_filter': True, 'is_sort_key': True,
               'primary_key': True},
        'project_id': {'allow_post': True, 'allow_put': False,
                       'required_by_policy': True,
                       'validate': {
                           'type:string':
                               db_const.PROJECT_ID_FIELD_SIZE},
                       'is_filter': True, 'is_sort_key': True,
                       'is_visible': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': db_const.NAME_FIELD_SIZE},
                 'is_filter': True, 'is_sort_key': True,
                 'default': '', 'is_visible': True},
        'resource_type': {'allow_post': True, 'allow_put': False,
                          'required_by_policy': True,
                          'validate': {
                              'type:string':
                                  db_const.RESOURCE_TYPE_FIELD_SIZE},
                          'is_filter': True, 'is_sort_key': True,
                          'is_visible': True},
        'resource_id': {'allow_post': True, 'allow_put': False,
                        'validate': {'type:uuid_or_none': None},
                        'is_filter': True, 'is_sort_key': True,
                        'default': None, 'is_visible': True},
        'event': {'allow_post': True, 'allow_put': False,
                  'validate': {'type:values': LOG_EVENTS},
                  'is_filter': True,
                  'default': ALL_EVENT, 'is_visible': True},
        'target_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:uuid_or_none': None},
                      'is_filter': True, 'is_sort_key': True,
                      'default': None, 'is_visible': True},
        'enabled': {'allow_post': True, 'allow_put': True,
                    'is_visible': True, 'default': True,
                    'is_filter': True, 'is_sort_key': True,
                    'convert_to': converters.convert_to_boolean},
    },
    LOG_TYPES: {
        'type': {'allow_post': False, 'allow_put': False,
                 'is_visible': True}},
}

# The subresource attribute map for the extension.  This extension has only
# top level resources, not child resources, so this is set to an empty dict.
SUB_RESOURCE_ATTRIBUTE_MAP = {}

# The action map.
ACTION_MAP = {}

# The list of required extensions.
REQUIRED_EXTENSIONS = []

# The action status.
ACTION_STATUS = {}

# The list of optional extensions.
OPTIONAL_EXTENSIONS = ['security-group', 'fwaas_v2']



# Copyright 2016 OpenStack Foundation
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import itertools

from oslo_serialization import jsonutils

from neutron_lib import constants as const
from neutron_lib.db import constants as db_const
from neutron_lib.objects import common_types
from neutron_lib.tests import _base as test_base
from neutron_lib.tests import tools
from neutron_lib.utils import net


class TestField(object):

    def test_coerce_good_values(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual(out_val, self.field.coerce('obj', 'attr', in_val))

    def test_coerce_bad_values(self):
        for in_val in self.coerce_bad_values:
            self.assertRaises((TypeError, ValueError),
                              self.field.coerce, 'obj', 'attr', in_val)

    def test_to_primitive(self):
        for in_val, prim_val in self.to_primitive_values:
            self.assertEqual(prim_val, self.field.to_primitive('obj', 'attr',
                                                               in_val))

    def test_to_primitive_json_serializable(self):
        for in_val, _ in self.to_primitive_values:
            prim = self.field.to_primitive('obj', 'attr', in_val)
            jsencoded = jsonutils.dumps(prim)
            self.assertEqual(prim, jsonutils.loads(jsencoded))

    def test_from_primitive(self):
        class ObjectLikeThing(object):
            _context = 'context'

        for prim_val, out_val in self.from_primitive_values:
            from_prim = self.field.from_primitive(ObjectLikeThing, 'attr',
                                                  prim_val)
            self.assertEqual(out_val, from_prim)
            # ensure it's coercable for sanity
            self.field.coerce('obj', 'attr', from_prim)

    @abc.abstractmethod
    def test_stringify(self):
        '''This test should validate stringify() format for new field types.'''


class IPV6ModeEnumFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(IPV6ModeEnumFieldTest, self).setUp()
        self.field = common_types.IPV6ModeEnumField()
        self.coerce_good_values = [(mode, mode)
                                   for mode in const.IPV6_MODES]
        self.coerce_bad_values = ['6', 4, 'type', 'slaacc']
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("'%s'" % in_val, self.field.stringify(in_val))


class DscpMarkFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(DscpMarkFieldTest, self).setUp()
        self.field = common_types.DscpMarkField()
        self.coerce_good_values = [(val, val)
                                   for val in const.VALID_DSCP_MARKS]
        self.coerce_bad_values = ['6', 'str', [], {}, object()]
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("%s" % in_val, self.field.stringify(in_val))


class IPNetworkPrefixLenFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(IPNetworkPrefixLenFieldTest, self).setUp()
        self.field = common_types.IPNetworkPrefixLenField()
        self.coerce_good_values = [(x, x) for x in (0, 32, 128, 42)]
        self.coerce_bad_values = ['len', '1', 129, -1]
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("%s" % in_val, self.field.stringify(in_val))


class MACAddressFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(MACAddressFieldTest, self).setUp()
        self.field = common_types.MACAddressField()
        mac1 = tools.get_random_EUI()
        mac2 = tools.get_random_EUI()
        self.coerce_good_values = [(mac1, mac1), (mac2, mac2)]
        self.coerce_bad_values = [
            'XXXX', 'ypp', 'g3:vvv',
            # the field type is strict and does not allow to pass strings, even
            # if they represent a valid MAC address
            net.get_random_mac('fe:16:3e:00:00:00'.split(':')),
        ]
        self.to_primitive_values = ((a1, str(a2))
                                    for a1, a2 in self.coerce_good_values)
        self.from_primitive_values = ((a2, a1)
                                      for a1, a2 in self.to_primitive_values)

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual('%s' % in_val, self.field.stringify(in_val))


class IPNetworkFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(IPNetworkFieldTest, self).setUp()
        self.field = common_types.IPNetworkField()
        addrs = [
            tools.get_random_ip_network(version=ip_version)
            for ip_version in const.IP_ALLOWED_VERSIONS
        ]
        self.coerce_good_values = [(addr, addr) for addr in addrs]
        self.coerce_bad_values = [
            'ypp', 'g3:vvv',
            # the field type is strict and does not allow to pass strings, even
            # if they represent a valid IP network
            '10.0.0.0/24',
        ]
        self.to_primitive_values = ((a1, str(a2))
                                    for a1, a2 in self.coerce_good_values)
        self.from_primitive_values = ((a2, a1)
                                      for a1, a2 in self.to_primitive_values)

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual('%s' % in_val, self.field.stringify(in_val))


class IPVersionEnumFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(IPVersionEnumFieldTest, self).setUp()
        self.field = common_types.IPVersionEnumField()
        self.coerce_good_values = [(val, val)
                                   for val in const.IP_ALLOWED_VERSIONS]
        self.coerce_bad_values = [5, 0, -1, 'str']
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("%s" % in_val, self.field.stringify(in_val))


class FlowDirectionEnumFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(FlowDirectionEnumFieldTest, self).setUp()
        self.field = common_types.FlowDirectionEnumField()
        self.coerce_good_values = [(val, val)
                                   for val in const.VALID_DIRECTIONS]
        self.coerce_bad_values = ['test', '8', 10, []]
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("'%s'" % in_val, self.field.stringify(in_val))


class FlowDirectionAndAnyEnumFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(FlowDirectionAndAnyEnumFieldTest, self).setUp()
        self.field = common_types.FlowDirectionAndAnyEnumField()
        self.coerce_good_values = [
            (val, val) for val in const.VALID_DIRECTIONS_AND_ANY]
        self.coerce_bad_values = ['test', '8', 10, []]
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("'%s'" % in_val, self.field.stringify(in_val))


class PortRangesFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(PortRangesFieldTest, self).setUp()
        self.field = common_types.PortRangesField()
        self.coerce_good_values = [(val, val) for val in (
            '80:80', '80:90', '80', 80)]
        self.coerce_bad_values = ('x', 0, 99999, '99999',
                                  '9999:', ':9999',
                                  '99999:100000', '80:70')
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values


class DomainNameFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(DomainNameFieldTest, self).setUp()
        self.field = common_types.DomainNameField()
        self.coerce_good_values = [
            (val, val)
            for val in ('www.google.com', 'hostname', '1abc.com')
        ]
        self.coerce_bad_values = ['x' * (db_const.FQDN_FIELD_SIZE + 1), 10, []]
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("'%s'" % in_val, self.field.stringify(in_val))


class EtherTypeEnumFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(EtherTypeEnumFieldTest, self).setUp()
        self.field = common_types.EtherTypeEnumField()
        self.coerce_good_values = [(val, val)
                                   for val in const.VALID_ETHERTYPES]
        self.coerce_bad_values = ['IpV4', 8, 'str', 'ipv6']
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("'%s'" % in_val, self.field.stringify(in_val))


class IpProtocolEnumFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(IpProtocolEnumFieldTest, self).setUp()
        self.field = common_types.IpProtocolEnumField()
        self.coerce_good_values = [
            (val, val)
            for val in itertools.chain(
                const.IP_PROTOCOL_MAP.keys(),
                [str(v) for v in range(256)]
            )
        ]
        self.coerce_bad_values = ['test', 'Udp', 256]
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("'%s'" % in_val, self.field.stringify(in_val))


class UUIDFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(UUIDFieldTest, self).setUp()
        self.field = common_types.UUIDField()
        self.coerce_good_values = [
            ('f1d9cb3f-c263-45d3-907c-d12a9ef1629e',
                'f1d9cb3f-c263-45d3-907c-d12a9ef1629e'),
            ('7188f6637cbd4097a3b1d1bb7897c7c0',
                '7188f6637cbd4097a3b1d1bb7897c7c0')]
        self.coerce_bad_values = [
            'f1d9cb3f-c263-45d3-907c-d12a9ef16zzz',
            '7188f6637cbd4097a3b1d1bb7897']
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual('%s' % in_val, self.field.stringify(in_val))


class DictOfMiscValuesFieldTest(test_base.BaseTestCase, TestField):
    def setUp(self):
        super(DictOfMiscValuesFieldTest, self).setUp()
        self.field = common_types.DictOfMiscValues
        test_dict_1 = {'a': True,
                       'b': 1.23,
                       'c': ['1', 1.23, True],
                       'd': {'aa': 'zz'},
                       'e': '10.0.0.1'}
        test_dict_str = jsonutils.dumps(test_dict_1)
        self.coerce_good_values = [
            (test_dict_1, test_dict_1),
            (test_dict_str, test_dict_1)
        ]
        self.coerce_bad_values = [str(test_dict_1), '{"a":}']
        self.to_primitive_values = [
            (test_dict_1, test_dict_str)
        ]
        self.from_primitive_values = [
            (test_dict_str, test_dict_1)
        ]

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual(jsonutils.dumps(in_val),
                             self.field.stringify(in_val))


class NetworkSegmentRangeNetworkTypeEnumFieldTest(test_base.BaseTestCase,
                                                  TestField):
    def setUp(self):
        super(NetworkSegmentRangeNetworkTypeEnumFieldTest, self).setUp()
        self.field = common_types.NetworkSegmentRangeNetworkTypeEnumField()
        self.coerce_good_values = [(val, val)
                                   for val in [const.TYPE_VLAN,
                                               const.TYPE_VXLAN,
                                               const.TYPE_GRE,
                                               const.TYPE_GENEVE]]
        self.coerce_bad_values = [const.TYPE_FLAT, 'foo-network-type']
        self.to_primitive_values = self.coerce_good_values
        self.from_primitive_values = self.coerce_good_values

    def test_stringify(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual("'%s'" % in_val, self.field.stringify(in_val))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api.definitions import subnet as subnet_def
from neutron_lib import constants

USE_DEFAULT_SUBNETPOOL = 'use_default_subnetpool'

ALIAS = 'default-subnetpools'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Default Subnetpools'
API_PREFIX = ''
DESCRIPTION = 'Provides ability to mark and use a subnetpool as the default.'
UPDATED_TIMESTAMP = '2016-02-18T18:00:00-00:00'
RESOURCE_NAME = subnet_def.RESOURCE_NAME
COLLECTION_NAME = subnet_def.COLLECTION_NAME
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        USE_DEFAULT_SUBNETPOOL: {'allow_post': True,
                                 'allow_put': False,
                                 'default': False,
                                 'convert_to': converters.convert_to_boolean,
                                 'is_visible': False},
    },
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = [constants.SUBNET_ALLOCATION_EXT_ALIAS]
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



# Copyright 2018 Ericsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# trait prefixes to be used after CUSTOM_
TRAIT_PREFIX_VNIC_TYPE = 'VNIC_TYPE_'
TRAIT_PREFIX_PHYSNET = 'PHYSNET_'

# TODO(przszc): Remove resource classes constants with next major release of
# neutron-lib. There were only a few occurrences in neutron source code where
# those constants were used, and all of them were replaced with
# os-resource-classes lib.
# **DEPRECATED** resource classes
CLASS_NET_BW_EGRESS_KBPS = 'NET_BW_EGR_KILOBIT_PER_SEC'
CLASS_NET_BW_INGRESS_KBPS = 'NET_BW_IGR_KILOBIT_PER_SEC'

# Optionally reported inventory parameters. Mandatory parameters like 'total'
# are left out intentionally. See also:
# https://docs.openstack.org/api-ref/placement
#        /#update-resource-provider-inventory
INVENTORY_OPTIONS = set([
    'allocation_ratio',
    'max_unit',
    'min_unit',
    'reserved',
    'step_size',
    ])



# Copyright 2019 Ericsson
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

ALIAS = 'floatingip-autodelete-internal'
IS_SHIM_EXTENSION = True
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Autodelete floating IPs of internal networks'
DESCRIPTION = 'Autodelete unused floating IPs' \
              ' of networks being turned to router:external=False'
UPDATED_TIMESTAMP = '2019-01-24T00:00:00-00:00'
RESOURCE_ATTRIBUTE_MAP = {}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}
